@@ -6,28 +6,32 @@
 
 #include "fd.h"
 #include "term.h"
+#include "func.h"
 #include "funcno.h"
 #include "kanji.h"
 
 #include <time.h>
-#include <dirent.h>
+#include <sys/file.h>
 #include <sys/stat.h>
 
 extern int columns;
 extern int filepos;
 extern int mark;
 extern int fnameofs;
-extern int lnstat;
+extern int dispmode;
 extern int sorton;
 extern int chgorder;
 extern int stackdepth;
+extern int copypolicy;
 extern namelist filestack[];
 extern char fullpath[];
 extern char *archivefile;
 extern char *destpath;
+extern int writefs;
+extern int savehist;
 
 #ifndef	PAGER
-static void dump();
+static VOID dump();
 #endif
 static int cur_up();
 static int cur_down();
@@ -52,7 +56,10 @@ static int mark_find();
 static int push_file();
 static int pop_file();
 static int symlink_mode();
+static int filetype_mode();
+static int dotfile_mode();
 static int log_dir();
+static int log_top();
 static int view_file();
 static int edit_file();
 static int sort_dir();
@@ -89,97 +96,101 @@ char *findpattern = NULL;
 reg_t *findregexp = NULL;
 char **sh_history = NULL;
 bindtable bindlist[MAXBINDTABLE] = {
-	{K_UP,		CUR_UP,		-1},
-	{K_DOWN,	CUR_DOWN,	-1},
-	{K_RIGHT,	CUR_RIGHT,	-1},
-	{K_LEFT,	CUR_LEFT,	-1},
-	{K_PPAGE,	ROLL_UP,	-1},
-	{K_NPAGE,	ROLL_DOWN,	-1},
-	{K_F(1),	LOG_DIR,	-1},
-	{K_F(2),	EXECUTE_FILE,	-1},
-	{K_F(3),	COPY_FILE,	-1},
-	{K_F(4),	DELETE_FILE,	-1},
-	{K_F(5),	RENAME_FILE,	-1},
-	{K_F(6),	SORT_DIR,	-1},
-	{K_F(7),	FIND_FILE,	-1},
-	{K_F(8),	TREE_DIR,	-1},
-	{K_F(9),	EDIT_FILE,	-1},
-	{K_F(10),	UNPACK_FILE,	-1},
-	{K_F(11),	ATTR_FILE,	-1},
-	{K_F(12),	INFO_FILESYS,	-1},
-	{K_F(13),	MOVE_FILE,	-1},
-	{K_F(14),	DELETE_DIR,	-1},
-	{K_F(15),	MAKE_DIR,	-1},
-	{K_F(16),	EXECUTE_SH,	-1},
-	{K_F(17),	WRITE_DIR,	-1},
-	{K_F(18),	BACKUP_TAPE,	-1},
-	{K_F(19),	VIEW_FILE,	-1},
-	{K_F(20),	PACK_FILE,	-1},
+	{K_UP,		CUR_UP,		255},
+	{K_DOWN,	CUR_DOWN,	255},
+	{K_RIGHT,	CUR_RIGHT,	255},
+	{K_LEFT,	CUR_LEFT,	255},
+	{K_NPAGE,	ROLL_UP,	255},
+	{K_PPAGE,	ROLL_DOWN,	255},
+	{K_F(1),	LOG_DIR,	255},
+	{K_F(2),	EXECUTE_FILE,	255},
+	{K_F(3),	COPY_FILE,	255},
+	{K_F(4),	DELETE_FILE,	255},
+	{K_F(5),	RENAME_FILE,	255},
+	{K_F(6),	SORT_DIR,	255},
+	{K_F(7),	FIND_FILE,	255},
+	{K_F(8),	TREE_DIR,	255},
+	{K_F(9),	EDIT_FILE,	255},
+	{K_F(10),	UNPACK_FILE,	255},
+	{K_F(11),	ATTR_FILE,	255},
+	{K_F(12),	INFO_FILESYS,	255},
+	{K_F(13),	MOVE_FILE,	255},
+	{K_F(14),	DELETE_DIR,	255},
+	{K_F(15),	MAKE_DIR,	255},
+	{K_F(16),	EXECUTE_SH,	255},
+	{K_F(17),	WRITE_DIR,	255},
+	{K_F(18),	BACKUP_TAPE,	255},
+	{K_F(19),	VIEW_FILE,	255},
+	{K_F(20),	PACK_FILE,	255},
 	{CR,		LAUNCH_FILE,	IN_DIR},
-	{K_BS,		OUT_DIR,	-1},
-	{K_DC,		PUSH_FILE,	-1},
-	{K_IC,		POP_FILE,	-1},
-	{'\t',		MARK_FILE,	-1},
-	{ESC,		QUIT_SYSTEM,	-1},
+	{K_BS,		OUT_DIR,	255},
+	{K_DC,		PUSH_FILE,	255},
+	{K_IC,		POP_FILE,	255},
+	{'\t',		MARK_FILE,	255},
+	{ESC,		QUIT_SYSTEM,	255},
 
-	{'<',		CUR_TOP,	-1},
-	{'>',		CUR_BOTTOM,	-1},
-	{'1',		ONE_COLUMN,	-1},
-	{'2',		TWO_COLUMNS,	-1},
-	{'3',		THREE_COLUMNS,	-1},
-	{'5',		FIVE_COLUMNS,	-1},
-	{'(',		FNAME_RIGHT,	-1},
-	{')',		FNAME_LEFT,	-1},
-	{' ',		MARK_FILE2,	-1},
-	{'+',		MARK_ALL,	-1},
-	{'*',		MARK_FIND,	-1},
-	{']',		PUSH_FILE,	-1},
-	{'[',		POP_FILE,	-1},
-	{'?',		HELP_MESSAGE,	-1},
+	{'<',		CUR_TOP,	255},
+	{'>',		CUR_BOTTOM,	255},
+	{'1',		ONE_COLUMN,	255},
+	{'2',		TWO_COLUMNS,	255},
+	{'3',		THREE_COLUMNS,	255},
+	{'5',		FIVE_COLUMNS,	255},
+	{'(',		FNAME_RIGHT,	255},
+	{')',		FNAME_LEFT,	255},
+	{' ',		MARK_FILE2,	255},
+	{'+',		MARK_ALL,	255},
+	{'*',		MARK_FIND,	255},
+	{']',		PUSH_FILE,	255},
+	{'[',		POP_FILE,	255},
+	{'?',		HELP_MESSAGE,	255},
 
-	{'a',		ATTR_FILE,	-1},
-	{'b',		BACKUP_TAPE,	-1},
-	{'c',		COPY_FILE,	-1},
-	{'d',		DELETE_FILE,	-1},
-	{'e',		EDIT_FILE,	-1},
-	{'f',		FIND_FILE,	-1},
-	{'h',		EXECUTE_SH,	-1},
-	{'i',		INFO_FILESYS,	-1},
-	{'k',		MAKE_DIR,	-1},
-	{'l',		LOG_DIR,	-1},
-	{'m',		MOVE_FILE,	-1},
-	{'p',		PACK_FILE,	-1},
-	{'q',		QUIT_SYSTEM,	-1},
-	{'r',		RENAME_FILE,	-1},
-	{'s',		SORT_DIR,	-1},
-	{'t',		TREE_DIR,	-1},
-	{'u',		UNPACK_FILE,	-1},
-	{'v',		VIEW_FILE,	-1},
-	{'w',		WRITE_DIR,	-1},
-	{'x',		EXECUTE_FILE,	-1},
-	{'C',		COPY_TREE,	-1},
-	{'D',		DELETE_DIR,	-1},
-	{'F',		FIND_DIR,	-1},
-	{'L',		LOG_TREE,	-1},
-	{'M',		MOVE_TREE,	-1},
-	{'Q',		QUIT_SYSTEM,	-1},
-	{'S',		SYMLINK_MODE,	-1},
-	{'U',		UNPACK_TREE,	-1},
-	{CTRL_A,	CUR_TOP,	-1},
-	{CTRL_B,	CUR_LEFT,	-1},
-	{CTRL_C,	ROLL_UP,	-1},
-	{CTRL_E,	CUR_BOTTOM,	-1},
-	{CTRL_F,	CUR_RIGHT,	-1},
-	{CTRL_L,	REREAD_DIR,	-1},
-	{CTRL_N,	CUR_DOWN,	-1},
-	{CTRL_P,	CUR_UP,		-1},
-	{CTRL_R,	ROLL_DOWN,	-1},
-	{CTRL_V,	ROLL_UP,	-1},
-	{CTRL_Y,	ROLL_DOWN,	-1},
-	{-1,		NO_OPERATION,	-1}
+	{'a',		ATTR_FILE,	255},
+	{'b',		BACKUP_TAPE,	255},
+	{'c',		COPY_FILE,	255},
+	{'d',		DELETE_FILE,	255},
+	{'e',		EDIT_FILE,	255},
+	{'f',		FIND_FILE,	255},
+	{'h',		EXECUTE_SH,	255},
+	{'i',		INFO_FILESYS,	255},
+	{'k',		MAKE_DIR,	255},
+	{'l',		LOG_DIR,	255},
+	{'\\',		LOG_TOP,	255},
+	{'m',		MOVE_FILE,	255},
+	{'p',		PACK_FILE,	255},
+	{'q',		QUIT_SYSTEM,	255},
+	{'r',		RENAME_FILE,	255},
+	{'s',		SORT_DIR,	255},
+	{'t',		TREE_DIR,	255},
+	{'u',		UNPACK_FILE,	255},
+	{'v',		VIEW_FILE,	255},
+	{'w',		WRITE_DIR,	255},
+	{'x',		EXECUTE_FILE,	255},
+	{'C',		COPY_TREE,	255},
+	{'D',		DELETE_DIR,	255},
+	{'F',		FIND_DIR,	255},
+	{'H',		DOTFILE_MODE,	255},
+	{'L',		LOG_TREE,	255},
+	{'M',		MOVE_TREE,	255},
+	{'Q',		QUIT_SYSTEM,	255},
+	{'S',		SYMLINK_MODE,	255},
+	{'T',		FILETYPE_MODE,	255},
+	{'U',		UNPACK_TREE,	255},
+	{CTRL_A,	CUR_TOP,	255},
+	{CTRL_B,	CUR_LEFT,	255},
+	{CTRL_C,	ROLL_UP,	255},
+	{CTRL_E,	CUR_BOTTOM,	255},
+	{CTRL_F,	CUR_RIGHT,	255},
+	{CTRL_L,	REREAD_DIR,	255},
+	{CTRL_N,	CUR_DOWN,	255},
+	{CTRL_P,	CUR_UP,		255},
+	{CTRL_R,	ROLL_DOWN,	255},
+	{CTRL_V,	ROLL_UP,	255},
+	{CTRL_Y,	ROLL_DOWN,	255},
+	{-1,		NO_OPERATION,	255}
 };
 
 
+/*ARGSUSED*/
 static int cur_up(list, maxp)
 namelist *list;
 int *maxp;
@@ -189,6 +200,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int cur_down(list, maxp)
 namelist *list;
 int *maxp;
@@ -198,42 +210,57 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int cur_right(list, maxp)
 namelist *list;
 int *maxp;
 {
-	if (filepos + FILEPERLOW >= *maxp) return(0);
-	filepos += FILEPERLOW;
+	if (filepos + FILEPERLOW >= *maxp) {
+		if (filepos / FILEPERPAGE == (*maxp - 1) / FILEPERPAGE)
+			return(0);
+		else filepos = *maxp - 1;
+	}
+	else filepos += FILEPERLOW;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int cur_left(list, maxp)
 namelist *list;
 int *maxp;
 {
 	if (filepos - FILEPERLOW < 0) return(0);
-	filepos -= FILEPERLOW;
+	else filepos -= FILEPERLOW;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int roll_up(list, maxp)
 namelist *list;
 int *maxp;
 {
-	if (filepos + FILEPERPAGE >= *maxp) return(0);
-	filepos += FILEPERPAGE;
+	int i;
+
+	i = (filepos / FILEPERPAGE) * FILEPERPAGE;
+	if (i + FILEPERPAGE >= *maxp) return(warning_bell(list, maxp));
+	filepos = i + FILEPERPAGE;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int roll_down(list, maxp)
 namelist *list;
 int *maxp;
 {
-	if (filepos - FILEPERPAGE < 0) return(0);
-	filepos -= FILEPERPAGE;
+	int i;
+
+	i = (filepos / FILEPERPAGE) * FILEPERPAGE;
+	if (i - FILEPERPAGE < 0) return(warning_bell(list, maxp));
+	filepos = i - FILEPERPAGE;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int cur_top(list, maxp)
 namelist *list;
 int *maxp;
@@ -243,6 +270,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int cur_bottom(list, maxp)
 namelist *list;
 int *maxp;
@@ -252,6 +280,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int in_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -261,6 +290,7 @@ int *maxp;
 	return(5);
 }
 
+/*ARGSUSED*/
 static int out_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -270,6 +300,7 @@ int *maxp;
 	return(5);
 }
 
+/*ARGSUSED*/
 static int one_column(list, maxp)
 namelist *list;
 int *maxp;
@@ -278,6 +309,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int two_columns(list, maxp)
 namelist *list;
 int *maxp;
@@ -286,6 +318,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int three_columns(list, maxp)
 namelist *list;
 int *maxp;
@@ -294,6 +327,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int five_columns(list, maxp)
 namelist *list;
 int *maxp;
@@ -302,6 +336,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int fname_right(list, maxp)
 namelist *list;
 int *maxp;
@@ -311,18 +346,29 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int fname_left(list, maxp)
 namelist *list;
 int *maxp;
 {
 	int i;
 
-	i = (islink(&list[filepos])) ? 0 : calcwidth();
-	if (fnameofs >= strlen(list[filepos].name) - i) return(0);
+	if (islink(&list[filepos])) i = 0;
+	else {
+		i = calcwidth();
+		if (isdisptyp(dispmode)
+		&& ((list[filepos].st_mode & S_IFMT) == S_IFDIR
+		|| (list[filepos].st_mode & S_IFMT) == S_IFLNK
+		|| (list[filepos].st_mode & S_IFMT) == S_IFSOCK
+		|| (list[filepos].st_mode & S_IFMT) == S_IFIFO
+		|| access(list[filepos].name, X_OK) >= 0)) i--;
+	}
+	if (fnameofs + i >= strlen(list[filepos].name)) return(0);
 	fnameofs++;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int mark_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -336,6 +382,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int mark_file2(list, maxp)
 namelist *list;
 int *maxp;
@@ -345,6 +392,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int mark_all(list, maxp)
 namelist *list;
 int *maxp;
@@ -367,6 +415,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int mark_find(list, maxp)
 namelist *list;
 int *maxp;
@@ -382,7 +431,7 @@ int *maxp;
 		return(1);
 	}
 
-	cp = cnvregexp(wild);
+	cp = cnvregexp(wild, 1);
 	re = regexp_init(cp);
 	free(wild);
 	free(cp);
@@ -397,6 +446,7 @@ int *maxp;
 	return(3);
 }
 
+/*ARGSUSED*/
 static int push_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -414,6 +464,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int pop_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -429,53 +480,75 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int symlink_mode(list, maxp)
 namelist *list;
 int *maxp;
 {
-	lnstat = 1 - lnstat;
+	dispmode ^= F_SYMLINK;
 	return(4);
 }
 
+/*ARGSUSED*/
+static int filetype_mode(list, maxp)
+namelist *list;
+int *maxp;
+{
+	dispmode ^= F_FILETYPE;
+	return(3);
+}
+
+/*ARGSUSED*/
+static int dotfile_mode(list, maxp)
+namelist *list;
+int *maxp;
+{
+	dispmode ^= F_DOTFILE;
+	return(4);
+}
+
+/*ARGSUSED*/
 static int log_dir(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *path;
 
-	if (!(path = evalpath(inputstr2(LOGD_K, -1, NULL, NULL)))) return(1);
-	if (!strcmp(path, "?")) {
-		free(list[filepos].name);
-		list[filepos].name = path;
-		return(5);
-	}
-	if (*path && chdir2(path) < 0) {
+	if (!(path = inputstr2(LOGD_K, -1, NULL, NULL))
+	|| !*(path = evalpath(path))) return(1);
+	if (!chdir3(path)) {
 		warning(-1, path);
 		free(path);
 		return(1);
 	}
-	if (!strcmp(path, ".")) {
-		free(path);
-		path = getwd2();
-		strcpy(fullpath, path);
-	}
-	else {
-		if (findpattern) free(findpattern);
-		findpattern = NULL;
-	}
 	free(path);
 	free(list[filepos].name);
 	list[filepos].name = strdup2("..");
 	return(4);
 }
 
+/*ARGSUSED*/
+static int log_top(list, maxp)
+namelist *list;
+int *maxp;
+{
+	char *path;
+
+	path = strdup2("/");
+	if (chdir2(path) < 0) error(path);
+	if (findpattern) free(findpattern);
+	findpattern = NULL;
+	free(list[filepos].name);
+	list[filepos].name = path;
+	return(4);
+}
+
 #ifndef	PAGER
-static void dump(file)
+static VOID dump(file)
 char *file;
 {
 	FILE *fp;
-	char *buf;
-	u_char *prompt;
+	char *buf, *prompt;
 	int i;
 
 	if (!(fp = fopen(file, "r"))) {
@@ -484,16 +557,19 @@ char *file;
 	}
 	putterms(t_clear);
 	tflush();
+	buf = NEXT_K;
+	i = strlen(file);
+	if (i + strlen(buf) > n_lastcolumn) i = n_lastcolumn - strlen(buf);
+	prompt = (char *)malloc2(i + strlen(buf) + 1);
+	strncpy3(prompt, file, i, 0);
+	strcat(prompt, buf);
 	buf = (char *)malloc2(n_column + 2);
-	prompt = (u_char *)malloc2(strlen(file) + sizeof(NEXT_K));
-	strcpy((char *)prompt, file);
-	strcat((char *)prompt, NEXT_K);
 
 	i = 0;
 	while (fgets(buf, n_column + 1, fp)) {
 		locate(0, i);
 		putterm(l_clear);
-		cputs(buf);
+		kanjiputs(buf);
 		if (++i >= n_line - 1) {
 			i = 0;
 			if (!yesno(prompt)) break;
@@ -512,6 +588,7 @@ char *file;
 }
 #endif
 
+/*ARGSUSED*/
 static int view_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -533,6 +610,7 @@ int *maxp;
 	return(2);
 }
 
+/*ARGSUSED*/
 static int edit_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -546,12 +624,13 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int sort_dir(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *str[5];
-	int i, tmp1, tmp2, val[5], *dup;
+	int i, tmp1, tmp2, val[5], *dupl;
 
 	str[0] = ONAME_K;
 	str[1] = OEXT_K;
@@ -567,7 +646,7 @@ int *maxp;
 	tmp2 = sorton & ~7;
 	i = (tmp1) ? 5 : 4;
 	if (!tmp1) tmp1 = val[0];
-	if (selectstr(&tmp1, i, 0, (u_char *)str, val) == ESC) return(1);
+	if (selectstr(&tmp1, i, 0, str, val) == ESC) return(1);
 
 	if (!tmp1) {
 		sorton = 0;
@@ -578,25 +657,26 @@ int *maxp;
 		str[1] = ODEC_K;
 		val[0] = 0;
 		val[1] = 8;
-		if (selectstr(&tmp2, 2, 40, (u_char *)str, val) == ESC)
+		if (selectstr(&tmp2, 2, 48, str, val) == ESC)
 			return(1);
 		sorton = tmp1 + tmp2;
-		dup = (int *)malloc(*maxp * sizeof(int));
+		dupl = (int *)malloc2(*maxp * sizeof(int));
 		for (i = 0; i < *maxp; i++) {
-			dup[i] = list[i].ent;
+			dupl[i] = list[i].ent;
 			list[i].ent = i;
 		}
 		qsort(list, *maxp, sizeof(namelist), cmplist);
 		for (i = 0; i < *maxp; i++) {
 			tmp1 = list[i].ent;
-			list[i].ent = dup[tmp1];
+			list[i].ent = dupl[tmp1];
 		}
-		free(dup);
+		free(dupl);
 		chgorder = 1;
 	}
 	return(3);
 }
 
+/*ARGSUSED*/
 static int write_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -604,20 +684,22 @@ int *maxp;
 #if (WRITEFS >= 2)
 	return(warning_bell(list, maxp));
 #else
-	if (atoi2(getenv2("FD_WRITEFS")) >= 2 || findpattern)
-		return(warning_bell(list, maxp));
-	if (writablefs(".") <= 0) {
+	int i;
+
+	if (writefs >= 2 || findpattern) return(warning_bell(list, maxp));
+	if ((i = writablefs(".")) <= 0) {
 		warning(0, NOWRT_K);
 		return(1);
 	}
 	if (!yesno(WRTOK_K)) return(1);
 	if (underhome() <= 0 && !yesno(HOMOK_K)) return(1);
-	arrangedir(list, *maxp);
+	arrangedir(list, *maxp, i);
 	chgorder = 0;
-	return(3);
+	return(4);
 #endif
 }
 
+/*ARGSUSED*/
 static int reread_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -626,6 +708,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int help_message(list, maxp)
 namelist *list;
 int *maxp;
@@ -634,26 +717,31 @@ int *maxp;
 	return(3);
 }
 
+/*ARGSUSED*/
 static int quit_system(list, maxp)
 namelist *list;
 int *maxp;
 {
 	if (!archivefile && !yesno(QUIT_K)) return(1);
+	if (savehist > 0) savehistory(sh_history, HISTORYFILE);
 	return(-1);
 }
 
+/*ARGSUSED*/
 static int make_dir(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *path;
 
-	if (!(path = evalpath(inputstr2(MAKED_K, -1, NULL, NULL)))) return(1);
-	if (*path && mkdir2(path, 0777) < 0) warning(-1, path);
+	if (!(path = inputstr2(MAKED_K, -1, NULL, NULL))
+	|| !*(path = evalpath(path))) return(1);
+	if (mkdir2(path, 0777) < 0) warning(-1, path);
 	free(path);
 	return(4);
 }
 
+/*ARGSUSED*/
 static int copy_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -661,6 +749,7 @@ int *maxp;
 	return(copyfile(list, *maxp, 0));
 }
 
+/*ARGSUSED*/
 static int copy_tree(list, maxp)
 namelist *list;
 int *maxp;
@@ -668,6 +757,7 @@ int *maxp;
 	return(copyfile(list, *maxp, 1));
 }
 
+/*ARGSUSED*/
 static int move_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -675,6 +765,7 @@ int *maxp;
 	return(movefile(list, *maxp, 0));
 }
 
+/*ARGSUSED*/
 static int move_tree(list, maxp)
 namelist *list;
 int *maxp;
@@ -682,6 +773,7 @@ int *maxp;
 	return(movefile(list, *maxp, 1));
 }
 
+/*ARGSUSED*/
 static int rename_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -692,8 +784,9 @@ int *maxp;
 	if (!strcmp(list[filepos].name, ".")
 	|| !strcmp(list[filepos].name, "..")) return(warning_bell(list, maxp));
 	for (;;) {
-		if (!(file = evalpath(inputstr2(NEWNM_K, 0,
-			list[filepos].name, NULL)))) return(1);
+		if (!(file = inputstr2(NEWNM_K, 0, list[filepos].name, NULL)))
+			return(1);
+		file = evalpath(file);
 		if (lstat(file, &status) < 0) {
 			if (errno == ENOENT) break;
 			warning(-1, file);
@@ -715,12 +808,14 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int delete_file(list, maxp)
 namelist *list;
 int *maxp;
 {
 	int i;
 
+	copypolicy = 0;
 	if (mark > 0) {
 		if (!yesno(DELMK_K)) return(1);
 		filepos = applyfile(list, *maxp, unlink2, NULL);
@@ -732,10 +827,11 @@ int *maxp;
 			warning(-1, list[filepos].name);
 		if (!i) filepos++;
 	}
-	if (filepos >= *maxp) filepos -= 2;
+	if (filepos >= *maxp && (filepos -= 2) < 0) filepos = 0;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int delete_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -746,6 +842,7 @@ int *maxp;
 	|| !strcmp(list[filepos].name, ".")
 	|| !strcmp(list[filepos].name, "..")) return(warning_bell(list, maxp));
 	if (!yesno(DELDR_K, list[filepos].name)) return(1);
+	copypolicy = 0;
 	if (islink(&list[filepos])) {
 		if ((i = unlink2(list[filepos].name)) < 0)
 			warning(-1, list[filepos].name);
@@ -757,6 +854,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int find_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -778,6 +876,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int find_dir(list, maxp)
 namelist *list;
 int *maxp;
@@ -791,7 +890,7 @@ int *maxp;
 		return(1);
 	}
 	destpath = NULL;
-	cp = cnvregexp(wild);
+	cp = cnvregexp(wild, 1);
 	findregexp = regexp_init(cp);
 	free(wild);
 	free(cp);
@@ -812,40 +911,21 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int execute_sh(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *com;
 
-	if (!sh_history) sh_history = entryhist(NULL, NULL);
-	if (!(com = inputstr2("sh#", -1, NULL, sh_history))) return(1);
-	sh_history = entryhist(sh_history, com);
-	if (!*com) {
-		free(com);
-		if (!(com = getenv2("FD_SHELL"))) com = "/bin/sh";
-		putterms(t_end);
-		putterms(t_nokeypad);
-		tflush();
-		cooked2();
-		echo2();
-		nl2();
-		cputs(SHEXT_K);
-		system(com);
-		raw2();
-		noecho2();
-		nonl2();
-		putterms(t_keypad);
-		putterms(t_init);
-	}
-	else {
-		if (*com == '!') execinternal(com + 1);
-		else execmacro(com, list[filepos].name, list, *maxp, 0, 1);
-		free(com);
-	}
+	if (!(com = inputstr2("sh#", -1, NULL, &sh_history))) return(1);
+	if (*com) execmacro(com, list[filepos].name, list, *maxp, 0, 1);
+	else execshell();
+	free(com);
 	return(4);
 }
 
+/*ARGSUSED*/
 static int execute_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -853,16 +933,15 @@ int *maxp;
 	char *com, *dir;
 	int len;
 
-	if (!sh_history) sh_history = entryhist(NULL, NULL);
 	len = (access(list[filepos].name, X_OK) >= 0) ?
 		strlen(list[filepos].name) + 1 : 0;
-	if (!(com = inputstr2("sh#", len, list[filepos].name, sh_history)))
+	if (!(com = inputstr2("sh#", len, list[filepos].name, &sh_history)))
 		return(1);
 	if (!*com) {
+		execshell();
 		free(com);
-		return(1);
+		return(4);
 	}
-	sh_history = entryhist(sh_history, com);
 	if (archivefile && !(dir = tmpunpack(list, *maxp))) {
 		free(com);
 		return(1);
@@ -876,17 +955,16 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int launch_file(list, maxp)
 namelist *list;
 int *maxp;
 {
-	char *path;
-
-	if (archivefile || launcher(list[filepos].name) < 0)
-		return(view_file(list, maxp));
+	if (launcher(list, *maxp) < 0) return(view_file(list, maxp));
 	return(4);
 }
 
+/*ARGSUSED*/
 static int pack_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -894,7 +972,8 @@ int *maxp;
 	char *file;
 	int i;
 
-	if (!(file = evalpath(inputstr2(PACK_K, -1, NULL, NULL)))) return(1);
+	if (!(file = inputstr2(PACK_K, -1, NULL, NULL))
+	|| !*(file = evalpath(file))) return(1);
 	i = pack(file, list, *maxp);
 	free(file);
 	if (i < 0) {
@@ -904,6 +983,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int unpack_file(list, maxp)
 namelist *list;
 int *maxp;
@@ -918,6 +998,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int unpack_tree(list, maxp)
 namelist *list;
 int *maxp;
@@ -934,6 +1015,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int info_filesys(list, maxp)
 namelist *list;
 int *maxp;
@@ -942,8 +1024,7 @@ int *maxp;
 	int i;
 
 	if (!(path = inputstr2(FSDIR_K, -1, NULL, NULL))) return(1);
-	if (*path) path = evalpath(path);
-	else {
+	if (!*(path = evalpath(path))) {
 		free(path);
 		path = strdup2(".");
 	}
@@ -953,12 +1034,13 @@ int *maxp;
 	return(3);
 }
 
+/*ARGSUSED*/
 static int attr_file(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *str[2];
-	int i, flag, val[2];
+	int i, n, flag, val[2];
 
 	str[0] = CMODE_K;
 	str[1] = CDATE_K;
@@ -971,8 +1053,8 @@ int *maxp;
 		flag = 1;
 		locate(0, LINFO);
 		putterm(l_clear);
-		cputs(ATTRM_K);
-		if (selectstr(&flag, 2, 40, (u_char *)str, val) == ESC)
+		kanjiputs(ATTRM_K);
+		if (selectstr(&flag, 2, 35, str, val) == ESC)
 			return(1);
 	}
 	else {
@@ -984,8 +1066,8 @@ int *maxp;
 		flag = 3;
 	}
 
-	while ((i = inputattr(&list[i], flag)) < 0) warning(0, ILTMS_K);
-	if (!i) return(3);
+	while ((n = inputattr(&list[i], flag)) < 0) warning(0, ILTMS_K);
+	if (!n) return(3);
 
 	if (mark > 0) applyfile(list, *maxp, setattr, NULL);
 	else if (setattr(list[filepos].name) < 0)
@@ -993,13 +1075,14 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int tree_dir(list, maxp)
 namelist *list;
 int *maxp;
 {
 	char *path;
 
-	if (!(path = tree())) return(3);
+	if (!(path = tree(0))) return(3);
 	if (chdir2(path) < 0) {
 		warning(-1, path);
 		free(path);
@@ -1013,6 +1096,7 @@ int *maxp;
 	return(4);
 }
 
+/*ARGSUSED*/
 static int backup_tape(list, maxp)
 namelist *list;
 int *maxp;
@@ -1020,13 +1104,15 @@ int *maxp;
 	char *dev;
 	int i;
 
-	if (!(dev = evalpath(inputstr2(BKUP_K, 5, "/dev/", NULL)))) return(1);
+	if (!(dev = inputstr2(BKUP_K, 5, "/dev/", NULL))
+	|| !*(dev = evalpath(dev))) return(1);
 	i = backup(dev, list, *maxp);
 	free(dev);
 	if (i <= 0) return(1);
 	return(4);
 }
 
+/*ARGSUSED*/
 static int warning_bell(list, maxp)
 namelist *list;
 int *maxp;
@@ -1035,6 +1121,7 @@ int *maxp;
 	return(0);
 }
 
+/*ARGSUSED*/
 static int no_operation(list, maxp)
 namelist *list;
 int *maxp;
